#include <u.h>
#include <libc.h>
#include <bio.h>
#include <pool.h>
#include <thread.h>
#include "misc.h"
#include "torrent.h"
#include "comm.h"

extern char mypeerid[];

static char *
xfer(int from)
{
	int datasize;
	int n, offset;
	char *to = nil;
	
	datasize = XFERSIZE;
	to = emallocz(datasize,1);
	offset = 0;
	while((n = readn(from, to+offset, XFERSIZE)) > 0){
		if (n < XFERSIZE) 
			break;
		datasize = datasize + XFERSIZE;
		to = erealloc(to, datasize);
		offset = offset + XFERSIZE;
	}
	if(n < 0) {
		free(to);
		sysfatal("read failed: %r");
	}
	return to;
}

static long 
readnumber(char *data, int *offset, char stop)
{
	int val;
	long nr;

	val = 0;
	while ((nr = data[*offset]) != stop)
	{
		if (nr == 'e') 
			/* enf of metainfo */
			return 0;
		if (nr <= 0)
			sysfatal("eof in readnumber");
		val = 10 * val + (nr - 48);	
		(*offset)++;
	}
	(*offset)++;
	return val;
}

static int
getkey(char *data, int *offset)
{
	uint length;
	char *key;
	int keytype;
	
	if ((length = readnumber(data, offset, ':')) == 0) 
		return -1;
	key = emalloc((length+1)*sizeof(char));
	memmove(key,&data[*offset],length);
	key[length]='\0';
	*offset = *offset + length;
	dbgprint("key: %s\n", key);
	for (int i = 0; i < nkeys; i++) {
		if (strcmp(key, keystab[i]) == 0) {
			keytype=i;
			dbgprint("[tracker]key found: %s\n", keystab[i]);
			break;
		}
	}
	free(key);
	return keytype;
}

static int
getelement(int keytype, char *data, int *offset, Peer **peers)
{
	int length;
	int newkeytype;
	static int index = 0;
	int threadindex = POOLSIZE-1;

	switch(keytype){
	case BTinterval:
		if (data[*offset] != 'i'){
			print("Not a valid beencoded integer\n");
			return -1;
		}
		(*offset)++;
		length = readnumber(data, offset, 'e');
		/*store that somewhere, maybe make a tracker struct*/
		dbgprint("interval: %d\n", length);
		break;
	case BTpeers:
		if (data[*offset] != 'l'){
			/* assume it's binary model data and not an error */
			dbgprint("binary model data\n");
			if ((length = readnumber(data, offset, ':')) == -1) 
				return -1;
			/* ip and port on 4+2 bytes, network order */
			length = length / 6;
			for (int i=0; i<length; i++){
				/*
				pool of POOLSIZE peers max for now
				just skip till the end of data after that
				*/
				if(i>threadindex)
					(*offset)+=6;
				else{
					peers[index]->address = ipcharstostring(&data[*offset]);
					(*offset)+=4;
					peers[index]->port = (uchar)data[*offset] * 256;
					(*offset)++;
					peers[index]->port += (uchar)data[*offset];
					(*offset)++;
					dbgprint("peers[%d]: %s:%d\n",index,peers[index]->address,peers[index]->port);
					index++;
				}
			}
		}
		else{
			/* dict model data */
			(*offset)++;
			if (data[*offset] == 'e')
				/* should not be an error, just timeout */
				error("No peers for this torrent");
			while (data[*offset] != 'e'){
				if (data[*offset] != 'd'){
					print("Not a valid dict in peers list\n");
					return -1;
				}
				(*offset)++;
				/* ip, peer_id, port */
				for (int i=0; i<3; i++){
					if ((newkeytype = getkey(data, offset)) == -1)
						return -1;
					if ((getelement(newkeytype, data, offset, peers)) == -1)
						return -1;
				}
				index++;
				if (data[*offset] != 'e'){
					print("Bad ending for dict in peers list\n");
					return -1;
				}
				(*offset)++;
			}
			(*offset)++;
		}
		break;
	case BTpeeraddress:
		if ((length = readnumber(data, offset, ':')) == -1) 
			return -1;
		if(index <= threadindex){
			peers[index]->address = emalloc((length+1)*sizeof(char));
			memmove(peers[index]->address,&data[*offset],length);
			peers[index]->address[length]='\0';
		}
		*offset = *offset + length;
		break;
	case BTpeerid:
		if ((length = readnumber(data, offset, ':')) == -1) 
			return -1;
		if(index <= threadindex){
			peers[index]->id = emalloc((length+1)*sizeof(char));
			memmove(peers[index]->id,&data[*offset],length);
			peers[index]->id[length]='\0';
		}
		*offset = *offset + length;
		break;
	case BTpeerport:
		if (data[*offset] != 'i')
			return -1;
		(*offset)++;
		length = readnumber(data, offset, 'e');
		if(index > threadindex)
			break;
		peers[index]->port = length;
		break;
	case BTcomplete:
		if (data[*offset] != 'i')
			return -1;
		(*offset)++;
		/*todo: use that info */
		readnumber(data, offset, 'e');
		break;
	case BTincomplete:
		if (data[*offset] != 'i')
			return -1;
		(*offset)++;
		/*todo: use that info better */
		peers[index]->seeder = readnumber(data, offset, 'e');
		if (peers[index]->seeder == 0)
			peers[index]->seeder = 1;
		break;
	case BTmininterval:
		if (data[*offset] != 'i'){
			print("Not a valid beencoded integer\n");
			return -1;
		}
		(*offset)++;
		length = readnumber(data, offset, 'e');
		/*store that somewhere, maybe make a tracker struct*/
		dbgprint("min interval: %d\n", length);
		break;
	case BTdownloaded:
		if (data[*offset] != 'i'){
			print("Not a valid beencoded integer\n");
			return -1;
		}
		(*offset)++;
		length = readnumber(data, offset, 'e');
		/*use that info later*/
		dbgprint("downloaded: %d\n", length);	
		break;
	default:
		print("No match!\n");
		return -1;
	}
}

static Peer **
parsetrackerreply(char *reply, Torrent *tor)
{
	int keytype = 0;
	int *offset; 
	int poolsize=POOLSIZE;

	/* create poolsize threads for now */
	tor->peers = emalloc(poolsize * sizeof(Peer *));
	for (int i=0; i<poolsize; i++){
		tor->peers[i] = emalloc(sizeof(Peer));
		tor->peers[i]->address = nil;
	}
	if (reply[0] != 'd'){
		print("Not a valid tracker reply");
		free(reply);
		return;
	}
	offset = emalloc(sizeof(int));
	*offset = 1;
	for (;;) 
	{
		if ((keytype = getkey(reply, offset)) == -1){
			if (reply[*offset] == 'e') 
				break;
			free(offset);
			error("getkey in parsetrackerreply");
		}
		if ((getelement(keytype, reply, offset, tor->peers)) == -1){
			free(offset);
			error("getelement in parsetrackerreply");
		}
	}
	tor->peersnb = 0;
	while(tor->peersnb < poolsize && tor->peers[tor->peersnb]->address != nil){
		tor->peers[tor->peersnb]->busy = 0;
		tor->peersnb++;
	}
	
	return tor->peers;
}

void
calltracker(Torrent *tor, char *reqtype)
{
	int conn, ctlfd, from, n;
	char buf[128];
	char *mtpt = "/mnt/web";
	char *reply = nil;
	char *msg = nil;
	Peer **peers = nil;
	Peer *peer = nil;
	int notracker = 0;

	if(!notracker){
	msg = forgerequest(tor, reqtype);
	/* webfs dance */
	snprint(buf, sizeof buf, "%s/clone", mtpt);
	if((ctlfd = open(buf, ORDWR)) < 0)
		sysfatal("couldn't open %s: %r", buf);
	if((n = read(ctlfd, buf, sizeof buf-1)) < 0)
		sysfatal("reading clone: %r");
	if(n == 0)
		sysfatal("short read on clone");
	buf[n] = '\0';
	conn = atoi(buf);
	dbgprint("query: %s\n", msg);
	/* "actual" request */
	if(fprint(ctlfd, "url %s", msg) <= 0)
		sysfatal("get ctl write: %r");
	snprint(buf, sizeof buf, "%s/%d/body", mtpt, conn);

	/* "actual" read */
	if((from = open(buf, OREAD)) < 0)
		sysfatal("open %s: %r", buf);
	}
	else{
	if((from = open("/usr/glenda/tracker_reply", OREAD)) < 0)
		sysfatal("open %s: %r", buf);
	}	
	reply = xfer(from);
	dbgprint("tracker reply: %s\n", reply);

	peers = parsetrackerreply(reply, tor);

	for(int i=0; i<tor->peersnb; i++){
		peer = peers[i];
		/*check if tracker sent "127.0.0.1" because it's bound to 0.0.0.0
		and there's a peer running on the same host */
		if ((strcmp("127.0.0.1", peer->address) == 0)
		&& (strcmp(mypeerid, peer->id) != 0)){
			dbgprint("tracker says peer on 127.0.0.1\n");
			/* assume announce starts with "http://" and fqdn ends with ":",
			will do better later */
			free(peer->address);
			peer->address = smprint("%s", strtok(&msg[7], ":"));
			dbgprint("peer->address: %s\n",peer->address);
		}
	}
	free(reply);
	free(msg);
}

static int
readlenpref(int peerfd)
{
	int m, n, length;
	uchar buf[4];
	Ioproc *io;

	n = sizeof(buf);
	if ((m = readn(peerfd, buf, n)) != n){
		print("problem reading length prefix \n");
		print("m was: %d\n", m);
		return -1;
	}
	length = buf[0] * (int)pow(2,24) + buf[1] * (int)pow(2,16) + buf[2] * (int)pow(2,8) + buf[3];
	/* hah that won't work because buf[] are chars 
	length = buf[0]<<24 + buf[1]<<16 + buf[2]<<8 + buf[3];
	*/
	return length;
}

static int 
readmsg(int peerfd, char *data)
{
	int n, length;
	char id;
	char *payload;
	ulong index, begin;
	Ioproc *io;

	if ((length = readlenpref(peerfd)) < 0)
		return -1;
	dbgprint("length: %d \n", length);
	if (length == 0){
		dbgprint("Keepalive. \n");
		return 10;	
	}
	if ((n = readn(peerfd, &id, 1)) < 0){
		dbgprint("Problem when reading message id. \n");
		return -1;
	}
	dbgprint("that's the id: %d \n", id);
	switch(id){
	case BTchoke:
		dbgprint("Choke received \n");
		break;
	case BTunchoke:
		break;
	case BTinterested:
		break;
	case BTnotinterested:
		break;
	case BThave:
		break;
	case BTbitfield:
		length--;
		dbgprint("Bitfield payload: %d bytes to read.\n", length);
		if ((n = readn(peerfd, data, length)) != length) {
			dbgprint("Problem when reading bitfield\n");
			free(data);
			return -1;
		}
		break;
	case BTrequest:
		break;
	case BTpiece:
		length -= 9;
		dbgprint("payload: %d bytes \n", length);
		if ((n = readn(peerfd, &index, sizeof(int))) != sizeof(int)){
			dbgprint("Problem when reading index of piece\n");
			return -1;
		}
		index = hton(&index);
		dbgprint("index of piece: %d \n", index);
		if ((n = readn(peerfd, &begin, sizeof(int))) != sizeof(int)){
			dbgprint("Problem when reading begin of piece\n");
			return -1;
		}
		begin = hton(&begin);
		dbgprint("begin of piece: %d \n", begin);
		payload = emalloc(length);
		if ((n = readn(peerfd, payload, length)) != length) {
			dbgprint("Problem when reading piece payload");
			free(payload);
			return -1;
		}
		memmove(&data[begin], payload, length);
		free(payload);
		break;
	case BTcancel:
		break;
	case BTport:
		break;
	default:
		dbgprint("Unknown id, something went wrong.\n");
		return -1;
	}

	return id;
}

static int 
writemsg(int peerfd, int id, uchar *data)
{
	int size, pos;
	uchar *msg;
	ulong index, begin, length;

	switch(id){
	case BTchoke:
		break;
	case BTunchoke:
		break;
	case BTinterested:
		size = 5;
		msg = emalloc(size);
		msg[0]=0;
		msg[1]=0;
		msg[2]=0;
		msg[3]=1;
		msg[4]=2;
		dbgprint("sending: ");
		for (int i=0; i<size; i++)
			dbgprint("%d ",msg[i]);
		dbgprint("\n");
		if (write(peerfd, msg, size) != size) {
			print("remote side hung up");
			free(msg);
			return -1;
		}
		free(msg);
		break;
	case BTnotinterested:
		break;
	case BThave:
		size = 9;
		msg = emalloc(size);
		msg[0]=0;
		msg[1]=0;
		msg[2]=0;
		msg[3]=5;
		msg[4]=4;
		pos = 0;
		memmove(&index, &data[pos], sizeof(int));
		index = hton(&index);
		pos = 5;
		memmove(&msg[pos], &index, sizeof(int));
		dbgprint("sending: ");
		for (int i=0; i<size; i++)
			dbgprint("%d ",msg[i]);
		dbgprint("\n");
		if (write(peerfd, msg, size) != size){
			dbgprint("Problem when writing \"have\"\n");
			free(msg);
			return -1;
		}
		free(msg);
		break;
	case BTbitfield:
		break;
	case BTrequest:
		size = 17;
		msg = emalloc(size);
		msg[0]=0;
		msg[1]=0;
		msg[2]=0;
		msg[3]=13;
		msg[4]=6;
		pos = 0;
		memmove(&index, &data[pos], sizeof(int));
		index = hton(&index);
		pos = pos + sizeof(int);
		memmove(&begin, &data[pos], sizeof(int));
		begin = hton(&begin);
		pos = pos + sizeof(int);
		memmove(&length, &data[pos], sizeof(int));
		length = hton(&length);
		pos = 5;
		memmove(&msg[pos], &index, sizeof(int));
		pos = pos + sizeof(int);
		memmove(&msg[pos], &begin, sizeof(int));
		pos = pos + sizeof(int);
		memmove(&msg[pos], &length, sizeof(int));
		
		dbgprint("sending: ");
		for (int i=0; i<size; i++)
			dbgprint("%d ",msg[i]);
		dbgprint("\n");
	
		if (write(peerfd, msg, size) != size){
			print("Problem when writing request\n");
			free(msg);
			return -1;
		}
		free(msg);
		break;
	case BTpiece:
		break;
	case BTcancel:
		break;
	case BTport:
		break;
	default:
		dbgprint("Unknown id, something went wrong.\n");
		return -1;
	}
	
	return id;
}

static int
hello(Peer *peer, Torrent *tor, Channel *c)
{
	char port[PORTSIZE];
	int peerfd;
	uchar *handshake;
	int pos, n, m;
	int bitfieldsize = tor->piecesnb / 8 + 1;
	Ioproc *io;
	char chanmsg[1];

	sprint(port,"%d",peer->port);
	dbgprint("addr called: %s\n",netmkaddr(peer->address, "tcp", port));
	io = ioproc();
	//peerfd = iodial(io, netmkaddr(peer->address, "tcp", port), 0, 0, 0);
	peerfd = dial(netmkaddr(peer->address, "tcp", port), 0, 0, 0);
	if(peerfd < 0){
		fprint(2, "can't dial %s: %r\n", peer->address);
		threadexits("dialing");
	}
	chanmsg[0] = '0';
	send(c,chanmsg);

	/*
	<pstrlen><pstr><reserved><info_hash><peer_id>
	*/
	n = 1 + PSTRLEN + 8 + HASHSIZE + PEERIDLEN;
	handshake = emallocz(n, 1);
	handshake[0] = PSTRLEN;
	pos = 1;
	memmove(&handshake[pos], pstr, PSTRLEN);
	pos = 1 + PSTRLEN + 8;
	memmove(&handshake[pos], tor->infohash, HASHSIZE);
	pos = 1 + PSTRLEN + 8 + HASHSIZE;
	memmove(&handshake[pos], mypeerid, PEERIDLEN);

	if (write(peerfd, handshake, n) != n){
		print("remote side hung up");
		free(handshake);	
		return -1;
	}

	/*
	this happens with rtorrent sometimes, dunno why,
	so we just kill this thread instead of aborting
	*/
	if ((m = readn(peerfd, handshake, n)) <= 0){
		free(handshake);	
		dbgprint("remote side didn't send anything, terminating.\n");
		threadexits("remote side didn't send anything \n");
	}
	/*
	if ((m = readn(peerfd, handshake, n)) <= 0){
		free(handshake);	
		print("remote side didn't send anything \n");
		return -1;
	}
	*/

	dbgprint("m: %d\n", m);
	for (int i=0; i<m; i++)
		dbgprint("%c",handshake[i]);
	dbgprint("\n");

	/* read, in case there's a bitfield */
	peer->bitfield = emalloc(bitfieldsize);
	if (readmsg(peerfd, peer->bitfield) != BTbitfield)
		dbgprint("Was expecting a bitfield \n");

	while(1){
		/* Send "interested" */
		/* Needed for handshake to be complete? */
		if (writemsg(peerfd, BTinterested, nil) != BTinterested)
			print("Problem when sending \"interested\" \n");
	
		send(c,chanmsg);

		/* read unchoke */
		if (readmsg(peerfd, nil) == BTunchoke)
			break;
		dbgprint("Was expecting an unchoke\n");
	}
	
	free(handshake);	
	closeioproc(io);
	return peerfd;
}

static int
getpieces(int peerfd, Peer *peer, Torrent *tor, Channel *c)
{
	int pos, size, piecelength, reqnb, piecesnb;
	uchar *msg;
	ulong index, begin, length;
	char *data;
	int m;
	Piece *lister, *rimmer, *kryten;
	char chanmsg[1];

	piecelength = tor->piecelength;
	piecesnb = tor->piecesnb;
	data = emalloc(piecelength);
	msg = emalloc(3*sizeof(int));
	reqnb = piecelength / REQSIZE;
	chanmsg[0]='0';

	//for (;;){
	while(peer->piecesnb != 0){
		/*
		pick a piece at random amongst the ones left
		each thread has its own list
		*/
		lister = peer->pieceslist;
		dbgprint("pieces to get from this peer: \n");
		while (lister != nil){
			dbgprint("[ %d ] ", lister->index);
			lister = lister->next;
		}
		dbgprint("\n");
		while(peer->piecesnb != 0){
			dbgprint("pieces left: %d\n", peer->piecesnb);
			index = pickrand(0,peer->piecesnb-1);
			dbgprint("pickrand: %d\n", index);
			lister = peer->pieceslist;
			for (int i=0; i<index; i++)
				lister = lister->next;
			index = lister->index;
			dbgprint("corresponding index: %d\n", index);
			rimmer = tor->pieceslist;
			while ((rimmer != nil) && (rimmer->index != index))
				rimmer = rimmer->next;
			//should not happen, so remove when sure
			if (rimmer == nil)
				error("main list empty before peer's list");
			if ((rimmer->status & BThas) == 0)
				break;
			dbgprint("drop this piece\n");
			/* remove from the peer's list if we already got this piece */
			lister = peer->pieceslist;
			kryten = lister;
			while ((lister != nil) && (lister->index != index)){
				kryten = lister;
				lister = lister->next;
			}
			//should not happen either, so remove when sure
			if (lister == nil)
				error("peer's list empty whereas we checked before");
			if (lister == peer->pieceslist)
				peer->pieceslist = lister->next;
			kryten->next = lister->next;
			free(lister);
			peer->piecesnb--;
		}
		dbgprint("random pick: %d\n", index);
		
		/*
		should only happen when our last piece has already
		been written by another thread. I'm checking  
		peer->piecesnb way too often, there probably
		is a better algo overall.
		*/
		if(peer->piecesnb == 0)
			break;

		length = REQSIZE;
		if (index == piecesnb-1)
			reqnb = tor->lastpiece / REQSIZE + 1;

		for (int j=0; j<reqnb; j++){
			pos = 0;
			begin = length * j;
			if ((index == piecesnb-1) && (j == reqnb - 1))
				length = tor->lastpiece - (tor->lastpiece / REQSIZE)*REQSIZE;
			memmove(&msg[pos], &index, sizeof(int));
			pos = pos + sizeof(int);
			memmove(&msg[pos], &begin, sizeof(int));
			pos = pos + sizeof(int);
			memmove(&msg[pos], &length, sizeof(int));
			if (writemsg(peerfd, BTrequest, msg) != BTrequest) {
				print("Problem when sending \"request\" \n");	
				free(data);
				free(msg);
				return -1;	
			}
			send(c,chanmsg);
			while ((m = readmsg(peerfd, data)) != BTpiece){
				dbgprint("Was expecting a piece\n");
				if (m != 10){
					dbgprint("from readmsg: %d\n", m);
					dbgprint("Assume connection was closed by peer\n");
					sleep(5000);
					free(data);
					free(msg);
					return -1;
				}
				dbgprint("go on.\n");
			}
		}
		
		if (checkpiece(data, tor, index) < 0){
			print("checkpiece failed for #%ld\n", index);
			return -1;
		}
	
		writedata(index, data, tor, peer);

		/* send "have" 
		<len=0005><id=4><piece index>
		*/
		pos = 0;
		memmove(&msg[pos], &index, sizeof(int));
		if (writemsg(peerfd, BThave, msg) != BThave) {
			print("Problem when sending \"have\" \n");	
			free(data);
			free(msg);
			return -1;	
		}
	}

	free(msg);
	free(data);
	dbgprint("end of getpieces\n");
	return 1;
}

void
callpeer(Peer *peer, Torrent *tor, Channel *c)
{
	int peerfd;

	if ((peerfd = hello(peer, tor, c)) <=0)
		error("Problem during hello");
	preppeerspieces(tor, peer);
	getpieces(peerfd, peer, tor, c);
	//free(peer);
	dbgprint("end of callpeer\n");
}
